#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ctpuz.h"

#define NUM_ROTATIONS  24
#define FIELD_SIZE     60
#define MAX_PIECE_SIZE 10

#define X_PIECE 10
#define ASYMMETRIC_PIECE 6
#define NUM_PIECES 12

#define coord2bit(x,y,z)  ((x)*12+(y)*3+(z)+4)
#define shrink_subs(x)  ((x) >> 10)

typedef int coord[3];
typedef coord piece[MAX_PIECE_SIZE];

/* The compiled pieces are held in a linked list.
 * For each firstbit value rotations hold the rotated pieces.
 */
typedef struct compiled_piece {
  struct compiled_piece *next;
  void                  *asmcode[64][8];
  long long             *rots[32][8];
  long                  *rots_hi[32][8];
  char                  *code[3];
  int                    nr;
} compiled_piece;

typedef struct hashed_pos {
  long long pos;
  int solutions;
  int substates;
} hashed_pos;

hashed_pos *hashtable;
#define hashcode(pos,used) \
    (((pos>>4) ^ (pos >> 13) ^ (pos >> 23) ^ (pos >> 40) ^ used) & (HASH_SIZE-1))

static int map_pattern(long x) {
    return (x&5) | ((x >> 10) & 2);
}

piece ctpieces[NUM_PIECES] = {
  {
    {0,0,0}, {1,0,0}, {0,1,0}, {0,2,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {1,0,0}, {2,0,0}, {1,1,0}, {0,2,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {0,1,0}, {0,2,0}, {0,3,0}, {1,3,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {0,1,0}, {1,1,0}, {0,2,0}, {0,3,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {0,1,0}, {1,1,0}, {1,1,1}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {2,0,0}, {1,1,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {0,1,0}, {1,1,0}, {2,1,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {0,1,0}, {1,1,0}, {1,2,0}, {2,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {0,1,0}, {0,1,1}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {1,1,0}, {2,1,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {1,0,0}, {0,1,0}, {1,1,0}, {2,1,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {2,0,0}, {3,0,0}, {0,1,0}, {2,1,0}, {-1,-1,-1}
  }
};

compiled_piece compiled[NUM_PIECES];
compiled_piece asym;

compiled_piece reordered[NUM_PIECES];
compiled_piece retractor;
compiled_piece *first_piece;

#ifdef STATISTICS
static void dumpBoard(long long board, long long piece) {
  char bchar[4] = {'.', 'p', 'X', '*'};
  int z, y, x;
  for (z = 2; z >= 0; z--) {
    for (y = 3; y >= 0; y--) {
      for (x = 0; x < 5; x++) {
	int i = coord2bit(x,y,z);
	int b = (board >> i) & 1;
	int p = (piece >> i) & 1;
	printf("%c ", bchar[b*2+p]);
      }
      printf("   ");
      for (x = 0; x < 5; x++) {
	int i = coord2bit(x,y,z);
	printf(" %2d", i);
      }
      printf("\n");
    }
    printf("\n");
  }
}
#endif

void rotate(coord c, int rot, coord dest) {
  switch (rot / 4) {
  case 0:
    dest[0] = c[0];
    dest[1] = c[1];
    dest[2] = c[2];
    break;
  case 1:
    dest[0] = c[0];
    dest[1] = c[2];
    dest[2] = -c[1];
    break;
  case 2:
    dest[0] = c[1];
    dest[1] = c[0];
    dest[2] = -c[2];
    break;
  case 3:
    dest[0] = c[1];
    dest[1] = c[2];
    dest[2] = c[0];
    break;
  case 4:
    dest[0] = c[2];
    dest[1] = c[1];
    dest[2] = -c[0];
    break;
  case 5:
    dest[0] = c[2];
    dest[1] = c[0];
    dest[2] = c[1];
    break;
  }
  if (rot & 1) {
    dest[0] = -dest[0];
    dest[2] = -dest[2];
  }
  if (rot & 2) {
    dest[1] = -dest[1];
    dest[2] = -dest[2];
  }
}

long long rotate_and_shift(piece p, int rot, int shift)
{
  int i;
  long long bitvec = 0LL;

  for (i = 0; i < MAX_PIECE_SIZE; i++) {
    coord dest;
    if (p[i][0] == -1)
      break;
    rotate(p[i], rot, dest);
    dest[0] += shift % 5;
    dest[1] += (shift / 5) % 4;
    dest[2] += (shift / 20);
    if (dest[0] < 0 || dest[0] >= 5
	|| dest[1] < 0 || dest[1] >= 4
	|| dest[2] < 0 || dest[2] >= 3) {
      /* piece doesn't fit.. */
      return 0LL;
    }
    bitvec |= 1LL << coord2bit(dest[0], dest[1], dest[2]);
  }
  return bitvec;
}

long long mirror(long long rotation) {
  long long rot = 0LL;
  int x,y,z;

  for (x=0; x < 5; x++) {
    for (y = 0; y < 4; y++) {
      for (z = 0; z < 3; z++) {
	if (rotation & (1LL << coord2bit(x,y,z))) {
	  rot |= 1LL << coord2bit(4-x,  y,2-z);
	}
      }
    }
  }
  return rot;
}

int best_full_rotation(long long rotation, int axes) {
  int x,y,z;
  long long rot0, rot1, rot2, rot3;

  if (axes == 2) {
    /* calculate the four rotations of the piece, but with
     * reversed bit order.
     */
    rot0 = rot1 = rot2 = rot3 = 0LL;
    for (x=0; x < 5; x++) {
      for (y = 0; y < 4; y++) {
	for (z = 0; z < 3; z++) {
	  if (rotation & (1LL << coord2bit(x,y,z))) {
	    rot0 |= 1LL << (63-coord2bit(  x,  y,  z));
	    rot1 |= 1LL << (63-coord2bit(4-x,3-y,  z));
	    rot2 |= 1LL << (63-coord2bit(  x,3-y,2-z));
	    rot3 |= 1LL << (63-coord2bit(4-x,  y,2-z));
	  }
	}
      }
    }
    return rot0 >= rot1
      &&   rot0 >= rot2
      &&   rot0 >= rot3;
    
  } else {
    /* calculate the four rotations of the piece, but with
     * reversed bit order.
     */
    rot0 = rot1 = rot2 = rot3 = 0LL;
    for (x=0; x < 5; x++) {
      for (y = 0; y < 4; y++) {
	for (z = 0; z < 3; z++) {
	  if (rotation & (1LL << coord2bit(x,y,z))) {
	    rot0 |= 1LL << (63 - coord2bit(  x,  y,  z));
	    rot3 |= 1LL << (63 - coord2bit(4-x,  y,2-z));
	  }
	}
      }
    }
    /* We want the lowest bit of current rotation be higher than for any
     * other of the four rotation.  Since we reversed the bits, we want
     * the highest bit of rot0 be smaller than for any other rot.
     * We can use normal < operation.
     */
    return rot0 <= rot3;
  }
}

static int order[] = {
    7, 6, 5, 4, 3, 2, 1, 0, -1,
    6, 4, 2, 0, -1,
    5, 1, 4, 0, -1
};
static int start[] = {
    0,-1,-1,-1, 4,-1, 6, 7, -1,
    1,-1, 5,-1,-1,
    2,-1, 3,-1,-1
};
static int mult[] ={
    3,2,2,1,3,2,2,1
};

void compile(int pnr, int reduce_symm, compiled_piece *dest) {
  int rot, shift, i, j, pat;
  long long rotpiece;
  int numrot[64][8];
  int totalrot = 0;
  int size = 28*3*sizeof(long long) + 32*3*sizeof(long);
  long long rots[64][8][24*6];
  long long *prots;
  long *prots_hi;

  dest->nr = pnr;
  
  for (shift = 0; shift < 64; shift++)
      for (pat = 0; pat < 8; pat++)
	  numrot[shift][pat] = 0;

  for (rot = 0; rot < NUM_ROTATIONS; rot++) {
    for (shift = 0; shift < FIELD_SIZE; shift++) {
      rotpiece = rotate_and_shift(ctpieces[pnr], rot, shift);
      if (!rotpiece)
	continue;

      if (reduce_symm) {
	/* Only take this piece if it is better than the 
	 * other 3 rotations with respect to the whole cuboid.
	 *
	 * Better means 
	 *   1. the lowest bit nr is higher.
	 *   2. the value is bigger.
	 */
	if (!best_full_rotation(rotpiece, reduce_symm))
	  continue;
      }

      /* put piece at right position in array */
      for (i = 0; i < FIELD_SIZE; i++) {
	if ((rotpiece & 1LL << i))
	  break;
      }
      pat = map_pattern(rotpiece >> (i+1));

      for (j = 0; j < numrot[i][pat]; j++) {
	if (rots[i][pat][j] == rotpiece)
	  /* Rotation was already seen */
	  goto next_rotation;
      }

      rots[i][pat][numrot[i][pat]] = rotpiece;
      numrot[i][pat]++;
      totalrot++;
      if (i > 31)
	  size += sizeof(long) * mult[pat];
      else
	  size += sizeof(long long) * mult[pat];

      if (numrot[i][pat] > 24 * 6) {
	printf("Too many rotations for %d,%d!",pnr,i);
	exit(0);
      }

      /*printf("%2d,%2d,%2d: %16Lx\n", pnr, i, rot, rotpiece);*/
      /*  dumpBoard(0, rotpiece);*/
    }
  next_rotation:
  }

  printf("Piece %d, %d rotations (%d bytes)\n", pnr, totalrot, size);

  prots = malloc(size);
  for (i = 4; i < 32; i++) {
      for (pat = 0; pat < (int) (sizeof(order)/sizeof(order[0])); pat++) {
	  if (start[pat] >= 0)
	      dest->rots[i][start[pat]] = prots;
	  if (order[pat] >= 0) {
	      for (j = 0; j < numrot[i][order[pat]]; j++)
		  *prots++ = rots[i][order[pat]][j];
	  } else 
	      *prots++ = 1LL;
      }
  }

  prots_hi = (long *) prots;
  for (i = 32; i < 64; i++) {
      for (pat = 0; pat < (int) (sizeof(order)/sizeof(order[0])); pat++) {
	  if (start[pat] >= 0)
	      dest->rots_hi[i-32][start[pat]] = prots_hi;
	  if (order[pat] >= 0) {
	      for (j = 0; j < numrot[i][order[pat]]; j++)
		  *prots_hi++ = rots[i][order[pat]][j] >> 32;
	  } else 
	      *prots_hi++ = 0L;
      }
  }
}

int solutions;


int used;
long long rotpieces[NUM_PIECES];
#ifdef COUNT_PLACED
unsigned int placedP[NUM_PIECES];
#endif
#ifdef COUNT_SAVED
int hits = 0;
unsigned int hitsP[NUM_PIECES] = {0};
unsigned int savedP[NUM_PIECES] = {0};
long long int saved = 0;
#endif
int placed = 0;

#ifdef COUNT_STEPS
int ctr[20] = {0,};
#endif

/**
 * Ralf Kleberhoff's Einer-Loch
 */
static int hasHole(long long besetzt) {
  return (~besetzt &
	  ((besetzt | 044444444444444444444LL<<4) <<  1) &
	  ((besetzt | 011111111111111111111LL<<4) >>  1) &
	  ((besetzt | 070007000700070007000LL<<4) <<  3) &
	  ((besetzt | 000070007000700070007LL<<4) >>  3) &
	  ((besetzt | 077770000000000000000LL<<4) << 12) &
	  ((besetzt | 000000000000000007777LL<<4) >> 12)) != 0LL;
}

static int find_first(long w) {
    w = ~w & (w-1);
    w -= (w >> 1) & 0x55555555L;
    w = ((w >> 2) & 0x33333333L) + (w & 0x33333333L);
    w = ((w >> 4) + w) & 0x0f0f0f0fL;
    w = ((w >> 8) + w);
    w = ((w >> 16) + w) & 0x000000ffL;
    return w;
}

#if 0
static void dumpSolution(int pieces) {
  char bchar[] = "0123456789AB....";
  int z, y, x, i;
  for (z = 2; z >= 0; z--) {
    for (y = 3; y >= 0; y--) {
      for (x = 0; x < 5; x++) {
	long long mask = 1LL << coord2bit(x,y,z);
	for (i = 0; i < pieces; i++) {
	  if ((rotpieces[i] & mask))
	    break;
	}
	printf("%c ", i == pieces? '.' : bchar[i]);
      }
      printf("\n");
    }
    printf("\n");
  }
}
#endif

#ifdef STATISTICS
void statistics() {
#if defined(COUNT_PLACED) || defined(COUNT_SAVED) || defined(COUNT_STEPS)
    int i;
#endif
    printf("placed: %u ", placed);
#ifdef COUNT_PLACED
    for (i = 0; i < NUM_PIECES; i++) {
	printf("%u+", placedP[i]);
    }
#endif
    printf("\n");
#ifdef COUNT_SAVED
    printf("Hits: %d  splitted:", hits);
    for (i = 0; i < NUM_PIECES; i++) {
	printf("%u+", hitsP[i]);
    }
    printf("\n");
    printf("Saved: %lld  splitted:", saved);
    for (i = 0; i < NUM_PIECES; i++) {
	printf("%u, ", savedP[i]);
    }
    printf("\n");
#endif
    
#ifdef COUNT_STEPS
    printf("Counters: ");
    for (i = 0; i < 20; i++) {
	printf("%u, ", ctr[i]);
    }
    printf("\n");
#endif
}
#endif

extern void solve_asm(long long board, int firstbit, int depth);
extern void retract_piece();

/* This is the C version of the recursive solve method.  There is also
 * an assembler version that is used for the last rotations, where speed
 * is important.  The C version differs in that it keeps a hashtable and
 * that it can't handle rot_hi.
 */
void solve(long long board, int firstbit, int depth) {
  int pat;
  int solved = 1;
  int hc, oldplaced, oldsolutions;
  struct compiled_piece **ppiece;

  if ((long) board ==  -1L) {
      solve_asm(board, firstbit, depth);
      return;
  }

#ifdef COUNT_PLACED
  placedP[depth-1]++;
#endif
  placed++;

  if (hasHole(board))
    return;

  
  firstbit = find_first(~(long) board);
  pat = map_pattern((long) (board >> (firstbit+1)));

  ppiece = &first_piece;
  while (*ppiece != &retractor) {
    struct compiled_piece *piece;
    long long *rots;

    piece = *ppiece;
    *ppiece = piece->next;
    used ^= 1 << piece->nr;
    rots = piece->rots[firstbit][pat];

    solved = 0;
    while (!(*(long*)rots & 1)) {
      long long piecemask = *rots++;

      if ((board & piecemask) == 0LL) {

	long long nboard;

	oldsolutions = solutions;
	oldplaced = placed;
	rotpieces[depth] = piecemask;

	nboard = board | piecemask;
	hc = hashcode(nboard, used);
	if (hashtable[hc].pos == nboard) {
	  int subs;
	  solutions += hashtable[hc].solutions;

	  subs = hashtable[hc].substates;
	  placed += subs;
#ifdef COUNT_SAVED
	  saved += subs;
	  savedP[depth] += subs;
	  hitsP[depth]++;
	  hits++;
#endif
	} else {	  
	  if (depth < HASH_DEPTH) {
	    solve(nboard, firstbit+1, depth + 1);
	  } else {
#if NUM_PIECES == HASH_DEPTH + 1
	    dumpSolution(depth + 1);
	    solutions++;
#else
	    solve_asm(nboard, firstbit+1, depth + 1);
#endif
	  }

	  if (hashtable[hc].substates < placed - oldplaced)
	  {
	    hashtable[hc].pos = nboard;
	    hashtable[hc].solutions = solutions - oldsolutions;
	    hashtable[hc].substates = placed - oldplaced;
	  }
	}
#if 0
	if (solutions == oldsolutions && (placed - oldplaced) > 100000) {
	  dumpSolution(depth + 1);
	  printf("has no solutions but %d pieces placed\n", placed - oldplaced);
	}
#endif
#ifdef STATISTICS
#if STATISTICS >= 2
	if (depth < STATISTICS /* && solutions > oldsolutions*/) {
	  printf("%d: %2d %4d solutions ; %7d solutions total\n", 
		 depth, piece->nr, solutions - oldsolutions, solutions);
	  dumpBoard(board, piecemask);
	  statistics();
	}
#endif
#endif
      } else if (depth == 1) {
	  printf("WEEIIIRRRDDD");
	  dumpBoard(board,piecemask);
      }
    }
    used ^= 1 << piece->nr;
    *ppiece = piece;
    ppiece = &piece->next;
  }
}


void asm_32_test(char **asm_pptr, int jump_offset, long rot, int ljmp) {
    static const char code[] = {
	0xf7, 0xc2, 0, 0, 0, 0,
#ifdef P4_CODE
	0x2e,
#define JMPSTART 7
#else
#define JMPSTART 6
#endif
	0x74
    };

    memcpy(*asm_pptr, code, sizeof(code));
    /* patch piece */
    memcpy(*asm_pptr+2, &rot, 4);

    if (ljmp) {
	(*asm_pptr)[JMPSTART] = 0x0f;
	(*asm_pptr)[JMPSTART+1] = 0x84;
	memcpy(*asm_pptr+JMPSTART+2, &jump_offset, 4);
	*asm_pptr += JMPSTART+6;
#define ASM_32_TEST_LLEN (JMPSTART+6)
    } else {
	(*asm_pptr)[JMPSTART+1] = jump_offset;
	*asm_pptr += JMPSTART+2;
#define ASM_32_TEST_SLEN (JMPSTART+2)
    }
}
void asm_32_block(char **asm_pptr, int jump_offset, long rot, int ljmp) {
    extern void solve_asm_32();
    long offset;
    static const char code[] = {
	0x52, 0x81, 0xca, 0, 0, 0, 0,
	0xe8, 0, 0, 0, 0,
	0x5a, 0xeb
    };

    memcpy(*asm_pptr, code, sizeof(code));
    /* patch piece */
    memcpy(*asm_pptr+3, &rot, 4);
    /* patch call */
    offset = (void*) solve_asm_32 - (void*) (*asm_pptr+8+4);
    memcpy(*asm_pptr+8, &offset, 4);
	
    if (ljmp) {
	(*asm_pptr)[13] = 0xe9;
	memcpy(*asm_pptr+14, &jump_offset, 4);
	*asm_pptr += 18;
#define ASM_32_BLOCK_LLEN 18
    } else {
#ifdef NO_ALIGNING
	(*asm_pptr)[14] = jump_offset;
	*asm_pptr += 15;
#define ASM_32_BLOCK_SLEN 15
#else
	(*asm_pptr)[14] = jump_offset+1;
	(*asm_pptr)[15] = 0x90;
	*asm_pptr += 16;
#define ASM_32_BLOCK_SLEN 16
#endif
    }
}

void asm_64_test(char **asm_pptr, int jump_offset, long long rot, int ljmp) {
    asm_32_test(asm_pptr, jump_offset, rot >> 32, ljmp);
#define ASM_64_TEST_LLEN ASM_32_TEST_LLEN
#define ASM_64_TEST_SLEN ASM_32_TEST_SLEN
}
void asm_64_block(char **asm_pptr, int jump_offset, long long rot, int ljmp) {
    extern void solve_asm_64();
    extern void solve_asm_32();
    long offset;
    static const char code[] = {
	0xf7, 0xc3, 0, 0, 0, 0,
	0x75, 0x15,
	0x52, 0x53, 
	0x81, 0xca, 0, 0, 0, 0,
	0x81, 0xcb, 0, 0, 0, 0,
	0xe8, 0, 0, 0, 0,
	0x5b, 0x5a, 0xeb
    };

    memcpy(*asm_pptr, code, sizeof(code));
    /* patch piece */
    offset = (long) rot;
    memcpy(*asm_pptr+2, &offset, 4);
    memcpy(*asm_pptr+18, &offset, 4);
    offset = (long) (rot >> 32);
    memcpy(*asm_pptr+12, &offset, 4);
    /* patch call */
    offset = (void*) solve_asm_64 - (void*) (*asm_pptr+23+4);
    /* if only the most significant bits of low word of rots are
     * set, we know that we can switch to solve_asm_32
     */
    if (!((long) ~rot  & (long) -rot))
	offset = (void*) solve_asm_32 - (void*) (*asm_pptr+23+4);
	
    memcpy(*asm_pptr+23, &offset, 4);
	
    if (ljmp) {
	(*asm_pptr)[29] = 0xe9;
	memcpy(*asm_pptr+30, &jump_offset, 4);
	*asm_pptr += 34;
#define ASM_64_BLOCK_LLEN 34
    } else {
#ifdef NO_ALIGNING
	(*asm_pptr)[30] = jump_offset;
	*asm_pptr += 31;
#define ASM_64_BLOCK_SLEN 31
#else
	(*asm_pptr)[30] = jump_offset+1;
	(*asm_pptr)[31] = 0x90;
	*asm_pptr += 32;
#define ASM_64_BLOCK_SLEN 32
#endif
    }
}
void asm_next_rot(char **asm_pptr) {
    long offset;
    static const char code[] = {
	/*0x89, 0x08, */
	0x89, 0xc8, 0x8b, 0x09,
	/*0x09, 0xc9, 0x0f, 0x84, 0, 0, 0, 0,*/
	/*0x8b, 0x31, 0x89, 0x30, */
	0xff, 0x24, 0xb9
    };

    memcpy(*asm_pptr, code, sizeof(code));
    /* patch jump */
    /*
    offset = (void *) retract_piece - (void*) (*asm_pptr+8+4);
    memcpy(*asm_pptr+8, &offset, 4);
    */

    *asm_pptr += 7;
#define ASM_NEXT_ROT_LEN 7
}

void next_piece();

#define ASM_32_EARLY ((128 - ASM_32_TEST_SLEN) / ASM_32_BLOCK_SLEN)
#define ASM_32_SHORT ((127 - ASM_NEXT_ROT_LEN - 15) / ASM_32_BLOCK_SLEN)

void assemble_32(struct compiled_piece *dest, long *rots_hi, 
		 int bit, int block) {
    int len, beglen, sstart;
    int i,j, pad1, pad2;
    long *p;
    char * assembler_code;

    len = 0;
    for (p = rots_hi; *p != 0L; p++)
	len++;

    if (len == 0) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots_hi[bit-32][j] == rots_hi)
		dest->asmcode[bit][j] = (void *) next_piece;
	}
	dest->code[block] = NULL;
	return;
    }

    if (len > ASM_32_SHORT)
	sstart = len - ASM_32_SHORT;
    else
	sstart = 0;

    if (len > ASM_32_EARLY + ASM_32_SHORT)
	beglen = ASM_32_EARLY;
    else
	beglen = sstart;

    pad1 = -(ASM_32_BLOCK_SLEN * beglen) & 15;
    pad2 = -(ASM_32_TEST_SLEN * (beglen + len - sstart)
	     + ASM_32_TEST_LLEN * (sstart - beglen)
	     + ASM_NEXT_ROT_LEN) & 15;

    dest->code[block] = assembler_code = malloc
	((sstart - beglen) * (ASM_32_TEST_LLEN + ASM_32_BLOCK_LLEN)
	 + (beglen + len - sstart) *  (ASM_32_TEST_SLEN + ASM_32_BLOCK_SLEN)
	 + ASM_NEXT_ROT_LEN + pad1 + pad2 + 15);
    assembler_code += -(long)assembler_code & 15;

    for (i = 0; i < beglen; i++) {
	asm_32_block(&assembler_code, 
		     ASM_32_BLOCK_SLEN * (beglen-i-1)
		     + ASM_32_TEST_SLEN * (i+1)
		     + pad1, 
		     rots_hi[i], 0);
    }
    memset(assembler_code, 0x90, pad1);
    assembler_code += pad1;
    for (i = 0; i < beglen; i++) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots_hi[bit-32][j] == &rots_hi[i])
		dest->asmcode[bit][j] = assembler_code;
	}
	asm_32_test(&assembler_code,
		    - (ASM_32_BLOCK_SLEN * (beglen-i) 
		       + pad1
		       + ASM_32_TEST_SLEN * (i+1)), 
		    rots_hi[i], 0);
    }

    for (i = beglen; i < sstart; i++) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots_hi[bit-32][j] == &rots_hi[i])
		dest->asmcode[bit][j] = assembler_code;
	}
	asm_32_test(&assembler_code, 
		    ASM_32_BLOCK_SLEN * ASM_32_SHORT
		    + ASM_32_BLOCK_LLEN * (i - beglen)
		    + ASM_NEXT_ROT_LEN + pad2
		    + ASM_32_TEST_SLEN * ASM_32_SHORT
		    + ASM_32_TEST_LLEN * (sstart - i - 1),
		    rots_hi[i], 1);
    }
    for (i = sstart; i < len; i++) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots_hi[bit-32][j] == &rots_hi[i])
		dest->asmcode[bit][j] = assembler_code;
	}
	asm_32_test(&assembler_code, 
		    ASM_32_BLOCK_SLEN * (i - sstart)
		    + ASM_NEXT_ROT_LEN + pad2
		    + ASM_32_TEST_SLEN * (len - i - 1),
		    rots_hi[i], 0);
    }

    for (j = 0; j < 8; j++) {
	if (dest->rots_hi[bit-32][j] == &rots_hi[len])
	    dest->asmcode[bit][j] = (void *) next_piece;
    }
    asm_next_rot(&assembler_code);
    memset(assembler_code, 0x90, pad2);
    assembler_code += pad2;

    for (i = sstart; i < len; i++) {
	asm_32_block(&assembler_code, 
		     - (ASM_32_TEST_SLEN * (len - i - 1)
			+ ASM_NEXT_ROT_LEN + pad2
			+ ASM_32_BLOCK_SLEN * (i + 1 - sstart)),
		     rots_hi[i], 0);
    }
    
    for (i = beglen; i < sstart; i++) {
	asm_32_block(&assembler_code, 
		     - (ASM_32_TEST_LLEN * (sstart - i - 1)
			+ ASM_32_TEST_SLEN * ASM_32_SHORT
			+ ASM_NEXT_ROT_LEN + pad2
			+ ASM_32_BLOCK_SLEN * ASM_32_SHORT
			+ ASM_32_BLOCK_LLEN * (i + 1 - beglen)),
		     rots_hi[i], 1);
	
    }
}

#define ASM_64_EARLY ((128 - ASM_64_TEST_SLEN) / ASM_64_BLOCK_SLEN)
#define ASM_64_SHORT ((127 - ASM_NEXT_ROT_LEN) / ASM_64_BLOCK_SLEN)

void assemble_64(struct compiled_piece *dest, long long *rots, 
		 int bit, int block) {
    int len, beglen, sstart;
    int i,j, pad1, pad2;
    long long *p;
    char * assembler_code;

    len = 0;
    for (p = rots; *p != 1LL; p++)
	len++;

    if (len == 0) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots[bit][j] == rots)
		dest->asmcode[bit][j] = (void *) next_piece;
	}
	dest->code[block] = NULL;
	return;
    }

    if (len > ASM_64_SHORT)
	sstart = len - ASM_64_SHORT;
    else
	sstart = 0;

    if (len > ASM_64_EARLY + ASM_64_SHORT)
	beglen = ASM_64_EARLY;
    else
	beglen = sstart;

    pad1 = -(ASM_64_BLOCK_SLEN * beglen) & 15;
    pad2 = -(ASM_64_TEST_SLEN * (beglen + len - sstart)
	     + ASM_64_TEST_LLEN * (sstart - beglen)
	     + ASM_NEXT_ROT_LEN) & 15;

    dest->code[block] = assembler_code = malloc
	((sstart - beglen) * (ASM_64_TEST_LLEN + ASM_64_BLOCK_LLEN)
	 + (beglen + len - sstart) *  (ASM_64_TEST_SLEN + ASM_64_BLOCK_SLEN)
	 + ASM_NEXT_ROT_LEN + pad1 + pad2 + 15);
    assembler_code += -(long)assembler_code & 15;

    for (i = 0; i < beglen; i++) {
	asm_64_block(&assembler_code, 
		     ASM_64_BLOCK_SLEN * (beglen-i-1)
		     + pad1
		     + ASM_64_TEST_SLEN * (i+1), 
		     rots[i], 0);
    }
    memset(assembler_code, 0x90, pad1);
    assembler_code += pad1;
    for (i = 0; i < beglen; i++) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots[bit][j] == &rots[i])
		dest->asmcode[bit][j] = assembler_code;
	}
	asm_64_test(&assembler_code,
		    - (ASM_64_BLOCK_SLEN * (beglen-i) 
		       + pad1
		       + ASM_64_TEST_SLEN * (i+1)), 
		    rots[i], 0);
    }

    for (i = beglen; i < sstart; i++) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots[bit][j] == &rots[i])
		dest->asmcode[bit][j] = assembler_code;
	}
	asm_64_test(&assembler_code, 
		    ASM_64_BLOCK_SLEN * ASM_64_SHORT
		    + ASM_64_BLOCK_LLEN * (i - beglen)
		    + ASM_NEXT_ROT_LEN + pad2
		    + ASM_64_TEST_SLEN * ASM_64_SHORT
		    + ASM_64_TEST_LLEN * (sstart - i - 1),
		    rots[i], 1);
    }
    for (i = sstart; i < len; i++) {
	for (j = 0; j < 8; j++) {
	    if (dest->rots[bit][j] == &rots[i])
		dest->asmcode[bit][j] = assembler_code;
	}
	asm_64_test(&assembler_code, 
		    ASM_64_BLOCK_SLEN * (i - sstart)
		    + ASM_NEXT_ROT_LEN + pad2
		    + ASM_64_TEST_SLEN * (len - i - 1),
		    rots[i], 0);
    }

    for (j = 0; j < 8; j++) {
	if (dest->rots[bit][j] == &rots[len])
	    dest->asmcode[bit][j] = (void *) next_piece;
    }
    asm_next_rot(&assembler_code);
    memset(assembler_code, 0x90, pad2);
    assembler_code += pad2;

    for (i = sstart; i < len; i++) {
	asm_64_block(&assembler_code, 
		     - (ASM_64_TEST_SLEN * (len - i - 1)
			+ ASM_NEXT_ROT_LEN + pad2
			+ ASM_64_BLOCK_SLEN * (i + 1 - sstart)),
		     rots[i], 0);
    }
    
    for (i = beglen; i < sstart; i++) {
	asm_64_block(&assembler_code, 
		     - (ASM_64_TEST_LLEN * (sstart - i - 1)
			+ ASM_64_TEST_SLEN * ASM_64_SHORT
			+ ASM_NEXT_ROT_LEN + pad2
			+ ASM_64_BLOCK_SLEN * ASM_64_SHORT
			+ ASM_64_BLOCK_LLEN * (i + 1 - beglen)),
		     rots[i], 1);
	
    }
}

void filter_piece(long long board,
		  struct compiled_piece *piece, 
		  struct compiled_piece *dest) {
    int size = 28*3*sizeof(long long) + 32*3*sizeof(long);
    long long rots[64][8][24*6];
    int numrot[64][8];
    int i, j, pat;
    long long *prots;
    long *prots_hi;
    for (i = 4; i < 64; i++)
	for (j = 0; j < 8; j++)
	    numrot[i][j] = 0;

    for (i = 4; i < 32; i++) {
	prots = piece->rots[i][0];
	while (*prots != 1ULL) {
	    long long rot = *prots;
	    pat = map_pattern(rot >> (i+1));
	    if (!(rot & board)) {
		rots[i][pat][numrot[i][pat]] = rot;
		numrot[i][pat]++;
		size += sizeof(long long) * mult[pat];
	    }
	    prots++;
	}
    }
    for (i = 32; i < 64; i++) {
	prots_hi = piece->rots_hi[i-32][0];
	while (*prots_hi != 0LL) {
	    long long rot = (long long) *prots_hi << 32;
	    pat = map_pattern(rot >> (i+1));
	    if (!(rot & board)) {
		rots[i][pat][numrot[i][pat]] = rot;
		numrot[i][pat]++;
		size += sizeof(long) * mult[pat];
	    }
	    prots_hi++;
	}
    }
    
    dest->nr = piece->nr;

    prots = malloc(size);
    for (i = 4; i < 32; i++) {
	for (pat = 0; pat < (int) (sizeof(order)/sizeof(order[0])); pat++) {
	    if (start[pat] >= 0)
		dest->rots[i][start[pat]] = prots;
	    if (order[pat] >= 0) {
		for (j = 0; j < numrot[i][order[pat]]; j++)
		    *prots++ = rots[i][order[pat]][j];
	    } else 
		*prots++ = 1LL;
	}
    }
    prots_hi = (long *) prots;
    for (i = 32; i < 64; i++) {
	for (pat = 0; pat < (int) (sizeof(order)/sizeof(order[0])); pat++) {
	    if (start[pat] >= 0)
		dest->rots_hi[i-32][start[pat]] = prots_hi;
	    if (order[pat] >= 0) {
		for (j = 0; j < numrot[i][order[pat]]; j++)
		    *prots_hi++ = rots[i][order[pat]][j] >> 32;
	    } else 
		*prots_hi++ = 0L;
	}
    }

    for (i = 4; i < 32; i++) {
	j = 0;
	for (pat = 0; pat < (int) (sizeof(order)/sizeof(order[0])); pat++) {
	    assemble_64(dest, dest->rots[i][start[pat]], i, j);
	    while (order[pat] != -1)
		pat++;
	}
    }
    
    for (i = 32; i < 64; i++) {
	j = 0;
	for (pat = 0; pat < (int) (sizeof(order)/sizeof(order[0])); pat++) {
	    assemble_32(dest, dest->rots_hi[i-32][start[pat]], i, j);
	    while (order[pat] != -1)
		pat++;
	}
    }
}

long long filter(long long board, struct compiled_piece *piece_list) {
    struct compiled_piece *p, **last;
    int i;

    if (first_piece) {
	for (p = first_piece; p; p = p->next) {
	    free(p->rots[4][0]);
	    for (i = 0; i < 3; i++)
		if (p->code[i])
		    free(p->code[i]);
	}
    }

    i = 0;
    last = &first_piece;
    for (p = piece_list; p; p = p->next) {
	filter_piece(board, p, &reordered[i]);
	*last = &reordered[i];
	last = &reordered[i].next;
	i++;
    }
    *last = &retractor;
    printf("filtering done\n");
    return board;
}

void placeFirst(long long rot) {
    int oldsolutions = solutions;
    long long board;

    /* The first piece is the x piece.  We must distinguish whether that
     * piece is in a symmetric position or not.  If it is symmetric we
     * need to use the asymmetric_piece to remove symmetric solutions.
     *
     * We also hack the used vector so that the hashtable isn't shared
     * between symmetric and asymmetric cases.
     */
    if (mirror(rot) == rot) {
	used = 1 << X_PIECE;
	compiled[ASYMMETRIC_PIECE-1].next = &asym;
    } else {
	used = 0;
	compiled[ASYMMETRIC_PIECE-1].next = &compiled[ASYMMETRIC_PIECE];
    }

    board = filter(15ULL | rot, &compiled[0]);
    rotpieces[0] = board;

    solve(board, coord2bit(0,0,0), 1);
#ifdef STATISTICS
    printf("0: %2d %4d solutions ; %7d solutions total\n", 
	   X_PIECE, solutions - oldsolutions, solutions);
    dumpBoard(15ULL, board);
    statistics();
#endif
}

int main() {
    int i, j;
    /* compile all pieces.  Remove all symmetries for the X_PIECE */
    for (i = 0 ; i< NUM_PIECES; i++) {
	compile(i, i == X_PIECE ? 2 : 0, &compiled[i]);
    }
    
    /* link the compiled pieces together */
    for (i = 0 ; i< NUM_PIECES-1; i++)
	compiled[i].next = &compiled[i+1];
    compiled[NUM_PIECES-1].next = NULL;

    /* remove X_PIECE from the list, as it is placed first. */
    compiled[X_PIECE-1].next = &compiled[X_PIECE+1];

    /* compile ASYMMETRIC_PIECE again with removed symmetries.  It
     * is used when the X_PIECE is symmetric.
     */
    compile(ASYMMETRIC_PIECE, 1, &asym);
    /* Set asym.next to right value. */
    asym.next = &compiled[ASYMMETRIC_PIECE + 1];
    
    printf("Rotation table built\n");
    
    first_piece = NULL;
    solutions = 0;
    placed = 0;
#ifdef COUNT_PLACED
    for (i = 0; i < NUM_PIECES; i++) {
	placedP[i] = 0;
    }
#endif
    for (i = 0; i < 64; i++) 
	for (j = 0; j < 8; j++)
	    retractor.asmcode[i][j] = retract_piece;
    retractor.next = NULL;
    
    hashtable = malloc(HASH_SIZE * sizeof(*hashtable));
    printf("Hashtable allocated\n");
    
    /* place x piece */
    for (i = 4; i < 32; i++) {
	long long *rots = compiled[X_PIECE].rots[i][0];
	while (*rots != 1LL) {
	    placeFirst(*rots);
	    rots++;
	}
    }
    for (i = 32; i < 64; i++) {
	long *rots_hi = compiled[X_PIECE].rots_hi[i-32][0];
	while (*rots_hi != 0L) {
	    placeFirst(((long long) *rots_hi) << 32);
	    rots_hi++;
	}
    }
    printf("solutions: %d\n", solutions);
    return 0;
}
